import abc
from abc import ABC
from decimal import Decimal
from importlib import import_module

from money.exceptions import InvalidExchangeBackend, ExchangeBackendNotSet


class BaseBackend(ABC):
    """Abstract base class API for exchange backends."""

    @property
    @abc.abstractmethod
    def base(self):
        """Returns the base currency."""

    @abc.abstractmethod
    def rate(self, currency):
        """Returns quotation between the base and another currency."""

    def quotation(self, origin, target):
        """Returns quotation between two currencies (origin, target)."""

        a = self.rate(origin)
        b = self.rate(target)
        if a and b:
            return b / a
        return None


class SimpleBackend(BaseBackend):
    def __init__(self):
        self._base = None
        self._rates = {}

    @property
    def base(self):
        return self._base

    @base.setter
    def base(self, currency):
        self._base = currency

    def setrate(self, currency, rate):
        self._rates[currency] = rate

    def rate(self, currency):
        if currency == self.base:
            return Decimal(1)
        return self._rates.get(currency, None)


class ExchangeRates:
    def __init__(self):
        self._backend = None

    @property
    def backend(self):
        """Returns the current backend."""

        return self._backend

    @property
    def backend_name(self):
        """Returns the class name of the current backend."""

        return self._backend.__class__.__name__

    @backend.setter
    def backend(self, backend):
        """Sets the current backend."""

        if backend is None:
            self._backend = None
            return

        if isinstance(backend, str):
            path, name = backend.rsplit('.', 1)
            module = import_module(path)
            backend = getattr(module, name)()
        elif isinstance(backend, type):
            backend = backend()

        if not isinstance(backend, BaseBackend):
            raise InvalidExchangeBackend()

        self._backend = backend

    def __getattr__(self, item):
        if self._backend is None:
            raise ExchangeBackendNotSet()
        return getattr(self._backend, item)

    def __setattr__(self, key, value):
        if key == '_backend' or key == 'backend':
            return super().__setattr__(key, value)
        if self._backend is None:
            raise ExchangeBackendNotSet()
        return setattr(self._backend, key, value)


xrates = ExchangeRates()



import re

from setuptools import setup, find_packages

with open('money/__init__.py', 'r', encoding='utf-8') as init:
    version = re.search('__version__ = [\'"]([^\'"]+)[\'"]', init.read()).group(1)

with open('README.md', 'r', encoding='utf-8') as readme:
    long_description = readme.read()

setup(
    name='money-lib',
    version=version,
    author='Rui Pereira',
    author_email='r4g3baby@gmail.com',
    license='MIT',
    description='Python 3 money lib with decimal precision and currency exchange support.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/R4G3_BABY/money-lib',
    packages=find_packages(exclude=['tests*']),
    install_requires=['Babel>=2.5.0'],
    python_requires='>=3',
    classifiers=[
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers'
    ]
)



import re

from babel.numbers import get_currency_precision, get_currency_name, get_currency_symbol

from money.exceptions import InvalidCurrencyFormat

CURRENCY_REGEX = re.compile('^[A-Z]{3}$')


class Currency:
    """
    Represents a currency identified by its ISO 4217 code.

    Parameters
    ----------
    currency_code: str
        The ISO 4217 code of the currency

    Raises
    ------
    InvalidCurrencyFormat
        If the given currency_code isn't a valid ISO 4217 format
    """

    def __init__(self, currency_code: str):
        if not CURRENCY_REGEX.match(currency_code):
            raise InvalidCurrencyFormat(currency_code)

        self._currency_code = currency_code

    @property
    def currency_code(self) -> str:
        """Returns the ISO 4217 code of this currency."""

        return self._currency_code

    @property
    def precision(self) -> int:
        """Returns the precision of this currency."""

        return get_currency_precision(self._currency_code)

    def display_name(self, locale: str = 'en_US') -> str:
        """Returns the name used by the locale for this currency."""

        return get_currency_name(self._currency_code, locale=locale)

    def symbol(self, locale: str = 'en_US') -> str:
        """Returns the symbol used by the locale for this currency."""

        return get_currency_symbol(self._currency_code, locale=locale)

    def __repr__(self):
        return self._currency_code

    def __str__(self):
        return self._currency_code

    def __eq__(self, other):
        if isinstance(other, Currency):
            return other.currency_code == self._currency_code
        if isinstance(other, str):
            return other == self._currency_code
        return NotImplemented

    def __ne__(self, other):
        return not self == other



from decimal import Decimal, InvalidOperation, ROUND_HALF_UP

from babel.numbers import format_currency

from money.currency import Currency
from money.exceptions import InvalidOperandType, InvalidAmount, ExchangeBackendNotSet, ExchangeRateNotFound
from money.exchange import xrates


class Money:
    """Class representing a monetary amount."""

    _rounding_mode = ROUND_HALF_UP

    def __init__(self, amount, currency):
        if not isinstance(currency, Currency):
            currency = Currency(str(currency))

        try:
            self._amount = Decimal(str(amount))
        except InvalidOperation:
            raise InvalidAmount(amount)

        self._currency = currency

    @property
    def real(self) -> Decimal:
        """Returns the real amount (without rounding)."""

        return self._amount

    @property
    def amount(self) -> Decimal:
        """Returns the amount rounded to the correct number of decimal places for the currency."""

        decimal_precision = Decimal(str(1 / (10 ** self._currency.precision)).rstrip('0'))
        return self._amount.quantize(decimal_precision, rounding=Money._rounding_mode)

    @property
    def currency(self) -> Currency:
        """Returns the currency."""

        return self._currency

    def __repr__(self):
        return '{} {}'.format(self._currency, self.amount)

    def __str__(self):
        return self.format()

    def __hash__(self):
        return hash((self._amount, self._currency))

    def __lt__(self, other):
        if not isinstance(other, Money):
            raise InvalidOperandType(other, '<')

        other = other.to(self._currency)

        return self._amount < other.real

    def __le__(self, other):
        if not isinstance(other, Money):
            raise InvalidOperandType(other, '<=')

        other = other.to(self._currency)

        return self._amount <= other.real

    def __eq__(self, other):
        if not isinstance(other, Money):
            return False

        return self.amount == other.amount and self._currency.currency_code == other.currency.currency_code

    def __ne__(self, other):
        return not self == other

    def __ge__(self, other):
        if not isinstance(other, Money):
            raise InvalidOperandType(other, '>=')

        other = other.to(self._currency)

        return self._amount >= other.real

    def __gt__(self, other):
        if not isinstance(other, Money):
            raise InvalidOperandType(other, '>=')

        other = other.to(self._currency)

        return self._amount > other.real

    def __add__(self, other):
        other = self._parse_other(other)

        return self.__class__(self._amount + other, self._currency)

    def __radd__(self, other):
        other = self._parse_other(other)

        return self.__class__(other + self._amount, self._currency)

    def __sub__(self, other):
        other = self._parse_other(other)

        return self.__class__(self._amount - other, self._currency)

    def __rsub__(self, other):
        other = self._parse_other(other)

        return self.__class__(other - self._amount, self._currency)

    def __mul__(self, other):
        other = self._parse_other(other)

        return self.__class__(self._amount * other, self._currency)

    def __rmul__(self, other):
        other = self._parse_other(other)

        return self.__class__(other * self._amount, self._currency)

    def __truediv__(self, other, inverse=False):
        other = self._parse_other(other)

        if other == Decimal(0):
            raise ZeroDivisionError

        return self.__class__(self._amount / other, self._currency)

    def __rtruediv__(self, other):
        other = self._parse_other(other)

        if self._amount == Decimal(0):
            raise ZeroDivisionError

        return self.__class__(other / self._amount, self._currency)

    def __floordiv__(self, other):
        other = self._parse_other(other)

        if other == Decimal(0):
            raise ZeroDivisionError

        return self.__class__(self._amount // other, self._currency)

    def __rfloordiv__(self, other):
        other = self._parse_other(other)

        if self._amount == Decimal(0):
            raise ZeroDivisionError

        return self.__class__(other // self._amount, self._currency)

    def __mod__(self, other):
        other = self._parse_other(other)

        if other == Decimal(0):
            raise ZeroDivisionError

        return self.__class__(self._amount % other, self._currency)

    def __rmod__(self, other):
        other = self._parse_other(other)

        if self._amount == Decimal(0):
            raise ZeroDivisionError

        return self.__class__(other % self._amount, self._currency)

    def __neg__(self):
        return self.__class__(-self._amount, self._currency)

    def __pos__(self):
        return self.__class__(+self._amount, self._currency)

    def __abs__(self):
        return self.__class__(abs(self._amount), self._currency)

    def __int__(self):
        return int(self._amount)

    def __float__(self):
        return float(self._amount)

    def __bool__(self):
        return bool(self._amount)

    def __round__(self, n=0):
        return self.__class__(round(self._amount, n), self._currency)

    def to(self, currency):
        """Returns the equivalent money object in another currency."""

        if currency == self._currency:
            return self

        if xrates.backend is None:
            raise ExchangeBackendNotSet()

        if not isinstance(currency, Currency):
            currency = Currency(currency)

        rate = xrates.backend.quotation(self._currency.currency_code, currency.currency_code)
        if rate is None:
            raise ExchangeRateNotFound(xrates.backend_name, self._currency, currency)

        return self.__class__(self._amount * rate, currency)

    def format(self, locale: str = 'en_US') -> str:
        """Returns a string of the currency formatted for the specified locale."""

        return format_currency(self._amount, self.currency.currency_code, locale=locale)

    def _parse_other(self, other) -> Decimal:
        if isinstance(other, Money):
            return other.to(self._currency).real
        return Decimal(other)

    @classmethod
    def set_rounding_mode(cls, mode):
        cls._rounding_mode = mode
